/*****************************************************
 * Routines to get a move, validate a move,
 * apply a move to the board,
 * and detect the end of the game.
 *
 * Kenrick Mock 12/20/01
 *****************************************************/
#include <iostream>
#include <stdio.h>
#include "globals.h"
#include "board.h"
using namespace std;

/************************************************************
 * Constructor for Board
 * Just initializes the board to empty.
 * TODO: It would be useful to add a copy constructor
 * so a board could be initialized to a different board
 ************************************************************/
Board::Board()
{
  m_whoseturn = 'b';
  InitBoard(); 		/* Sets board pieces to initial config */
}


/************************************************************
 * ShowBoard
 * Displays the board in beautiful ASCII.
 * Requires a display that can show approximately 30 vertical
 * lines, so a 25 line screen is too small to show the whole
 * board without scrolling.
 ************************************************************/
void Board::ShowBoard()
{
 int i,j;
 char c;

 for (j=1; j<DIM; j++) {
	cout << "    /\\  /\\  /\\  /\\  /\\  /\\  /\\\n";
	if (j<10) cout << " " << j << " ";
   	else cout << j << " ";
	for (i=1; i<DIM; i+=2) {
		c=' ';
		if (m_board[i][j]!=EMPTY) c=m_board[i][j];
	   	cout << "/ " << c << "\\";
	}
	cout << endl;
	j++;
	cout << "   \\  /\\  /\\  /\\  /\\  /\\  /\\  /\n";
	if (j<DIM) {
		if (j<10) cout << " " << j << " ";
   		else cout << j << " ";
		for (i=2; i<DIM; i+=2) {
			c=' ';
			if (m_board[i][j]!=EMPTY) c=m_board[i][j];
			cout << " \\/" << c;
		}
		cout << " \\/\n";
	}
	else {
		cout << "    \\/  \\/  \\/  \\/  \\/  \\/  \\/\n";
		cout << "      2   4   6   8   10  12\n";
		cout << "     1  3   5   7   9   11  13\n";
	}
 }
}

/************************************************************
 * InitBoard
 * Initialize the board array.
 * number 0 = empty
 * w = white runner,  @ = white guard
 * b = black runner,  B = black guard
 ************************************************************/
void Board::InitBoard()
{
 int i,j;
 for (i=1; i<DIM; i++) 
	for (j=1; j<DIM; j++) 
		m_board[i][j]=EMPTY;
 for (i=3; i<13; i+=2) {
	m_board[i][1]=B_RUNNER;
	m_board[i][13]=W_RUNNER;
 }
 for (i=4; i<12; i+=2) {
	m_board[i][2]=B_GUARD;
	m_board[i][12]=W_GUARD;
 }
}


/*****************************************************
 * IsValid
 *
 * Returns 1 if the move is a valid one, 0 otherwise.
 * x1,y1 = source coordinates for move
 * x2,y2 = destination coordinates for move
 *****************************************************/
int Board::IsValid(int x1, int y1, int x2, int y2) 
{
 int i=0, temp=0;

 /* Invalid if any value is out of bounds */
 if ((x1<1) || (x1>=DIM) ||
     (y1<1) || (y1>=DIM) ||
     (x2<1) || (x2>=DIM) ||
     (y2<1) || (y2>=DIM)) return 0;

 /* Make sure there is a piece of the proper color at the source */
 if (m_whoseturn=='w') {
	if ((m_board[x1][y1]!=W_RUNNER) && (m_board[x1][y1]!=W_GUARD)) return 0;
 }
 else {
	if ((m_board[x1][y1]!=B_RUNNER) && (m_board[x1][y1]!=B_GUARD)) return 0;
 } 

 /* Check if move is valid */
 /* Move if this is a guard */
 if ((m_board[x1][y1]==W_GUARD) || (m_board[x1][y1]==B_GUARD)) {
	if (((x2 % 2)==1) || ((y2 % 2)==1)) return 0;   /* Dest must be even */
	/* Vertical move */
	if (x1==x2) {
		if (y1>y2) {	/* Moving up */
			for (i=y1-2; i>=y2; i-=2) {
				if (m_board[x1][i]!=EMPTY) return 0;
			}
			return 1;
		}
		else {		/* Moving down */
			for (i=y1+2; i<=y2; i+=2) {
				if (m_board[x1][i]!=EMPTY) return 0;
			}
			return 1;
		}	
	}
	/* Horizontal move */
	else if (y1==y2) {
		if (x1>x2) { 	/* Moving left */
			for (i=x1-2; i>=x2; i-=2) {
				if (m_board[i][y1]!=EMPTY) return 0;
			}
			return 1;
		}
		else {		/* Moving right */
			for (i=x1+2; i<=x2; i+=2) {
				if (m_board[i][y1]!=EMPTY) return 0;
			}
			return 1;
		}
	}
	else return 0;
 }

 /* Move if this is a white runner */
 else if (m_board[x1][y1]==W_RUNNER) {
	if (((x2 % 2)==0) || ((y2 % 2)==0)) return 0;   /* Dest must be odd */
	if (y1==y2) {		/* Side to side move */
		temp=x1-x2;
		if (temp<0) temp=temp*-1;
		if (temp>2) return 0;		/* Moved more than 1 square */
		if (m_board[x2][y2]==EMPTY) return 1;
	}
	/* Must be a upward move for white */
	if (x1!=x2) return 0;
	if (y1<y2) return 0;
	temp=y1-y2;
	if (temp<0) temp=temp*-1;
	if (temp>2) return 0;
	if (m_board[x2][y2]==EMPTY) return 1;
	else return 0;
 }

 /* Move if this is a black runner */
 else if (m_board[x1][y1]==B_RUNNER) {
	if (((x2 % 2)==0) || ((y2 % 2)==0)) return 0;   /* Dest must be odd */
	if (y1==y2) {		/* Side to side move */
		temp=x1-x2;
		if (temp<0) temp=temp*-1;
		if (temp>2) return 0;		/* Moved more than 1 square */
		if (m_board[x2][y2]==EMPTY) return 1;
	}
	/* Must be a downward move for white */
	if (x1!=x2) return 0;
	if (y1>y2) return 0;
	temp=y1-y2;
	if (temp<0) temp=temp*-1;
	if (temp>2) return 0;
	if (m_board[x2][y2]==EMPTY) return 1;
	else return 0;
 }

 return 1;
}

/***************************************************************************
 * RemoveRunner
 *
 * Returns 1 if the runner at the coordinates should be removed,
 *   0 otherwise.
 *   Assumes the input x and y are those for a valid runner.
 **************************************************************************/
int Board::RemoveRunner(int x, int y)
{
  int enemycount=0;
  char enemyGuard=W_GUARD;

  if (m_board[x][y]==W_RUNNER) enemyGuard=B_GUARD;
  if (x>1) {
	if (y>1) {
		if (m_board[x-1][y-1]==enemyGuard) enemycount++;
	}
	if (y<DIM-1) {
		if (m_board[x-1][y+1]==enemyGuard) enemycount++;
	}
  } 
  if (x<DIM-1) {
	if (y>1) {
		if (m_board[x+1][y-1]==enemyGuard) enemycount++;
	}
	if (y<DIM-1) {
		if (m_board[x+1][y+1]==enemyGuard) enemycount++;
	}
  } 

  if (enemycount>=2) {
	cout << "Runner at " << x << " " << y << " removed.\n";
	return 1;
  }
  else return 0;
}


/*****************************************************
 * ApplyMove
 *
 * Takes a valid move and applies it to the board.
 * This moves the actual piece on the board and
 * removes any pieces that may be surrounded by two
 * enemy guards.
 *****************************************************/
void Board::ApplyMove(int x1, int y1, int x2, int y2)
{
 int i;
 char c, enemyRunner=B_RUNNER;

 /* Move the piece on the board */
 m_board[x2][y2]=m_board[x1][y1];
 m_board[x1][y1]=EMPTY;

 /* Remove any runners that are touching two or more enemy guards */
 c = m_board[x2][y2];
 if ((c==B_RUNNER) || (c==W_RUNNER)) {
	/* If a runner moved, it is the only piece that can be removed */
	if (RemoveRunner(x2, y2)) {
		m_board[x2][y2]=EMPTY;
	}
 } 
 else {
	/* If a guard moved, we need to check around the guard to
           see if any touching runners can be removed */
	if (c==B_GUARD) enemyRunner=W_RUNNER;
	if (m_board[x2-1][y2-1]==enemyRunner) {
		if (RemoveRunner(x2-1, y2-1)) {
			m_board[x2-1][y2-1]=EMPTY;
		}
	}
	if (m_board[x2-1][y2+1]==enemyRunner) {
		if (RemoveRunner(x2-1, y2+1)) {
			m_board[x2-1][y2+1]=EMPTY;
		}
	}
	if (m_board[x2+1][y2-1]==enemyRunner) {
		if (RemoveRunner(x2+1, y2-1)) {
			m_board[x2+1][y2-1]=EMPTY;
		}
	}
	if (m_board[x2+1][y2+1]==enemyRunner) {
		if (RemoveRunner(x2+1, y2+1)) {
			m_board[x2+1][y2+1]=EMPTY;
		}
	}
 }
 // Change whose turn it is
 if (m_whoseturn=='b') m_whoseturn='w';
 else m_whoseturn='b';
}

/*****************************************************
 * GameOver
 *
 * Returns 1 if the game has ended, 0 otherwise.
 * Also prints out the winner.  
 *****************************************************/
int Board::GameOver()
{
 int w_runners=0, b_runners=0,i,j;
 
 for (i=1; i<DIM; i+=2) {
	if (m_board[i][1]==W_RUNNER) {
		cout << "wHITE WINS!\n";
		return 1;
	}
	if (m_board[i][DIM-1]==B_RUNNER) {
		cout << "BLACK WINS!\n";
		return 1;
	}
 }
 // Count up runners
 for (j=1; j<DIM; j+=2) {
	for (i=1; i<DIM; i+=2) {
		if (m_board[i][j]==W_RUNNER) w_runners++;
		else if (m_board[i][j]==B_RUNNER) b_runners++;
	}
 }

 if (w_runners==0) {
	cout << "BLACK WINS! No white runners left.\n";
	return 1;
 }
 if (b_runners==0) {
	cout << "WHITE WINS! No black runners left.\n";
	return 1;
 }
 return 0;
}



/*****************************************************
 * GetMove
 *
 * Gets a move from standard input.  The move must be
 * a valid one, or the user will be prompted to enter
 * the move again.
 *****************************************************/
void Board::GetMove(int &x1, int &y1, int &x2, int &y2)
{
  if (m_whoseturn=='w') cout << "\nWhite's move: ";
  else cout << "\nBlack's move: " ;
  do {
    cout << "Enter src and dest coords (e.g., x1 y1 x2 y2): ";
    fflush(stdout);
    cin >> x1 >> y1 >> x2 >> y2;
  } while (!IsValid(x1, y1, x2, y2));
}

